'''
Created on 2016/4/13

:author: hubo
'''
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.server.module import depend, ModuleNotification, call_api,api
import vlcp.service.sdn.ofpportmanager as ofpportmanager
import vlcp.service.sdn.ovsdbportmanager as ovsdbportmanager
import vlcp.service.kvdb.objectdb as objectdb
from vlcp.event.event import Event, withIndices, M_
from vlcp.event.runnable import RoutineContainer, RoutineException
from vlcp.config.config import defaultconfig
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.utils.networkmodel import PhysicalPort, LogicalPort, PhysicalPortSet, LogicalPortSet, LogicalNetwork, \
    PhysicalNetwork,SubNet,RouterPort,VRouter, \
    PhysicalNetworkMap
from vlcp.utils.flowupdater import FlowUpdater

import itertools
from functools import partial
from contextlib import closing, suppress
from vlcp.utils.exceptions import WalkKeyNotRetrieved
from vlcp.protocol.openflow.openflow import OpenflowConnectionStateEvent

@withIndices('datapathid', 'vhost', 'connection', 'logicalportchanged', 'physicalportchanged',
                                                    'logicalnetworkchanged', 'physicalnetworkchanged')
class DataObjectChanged(Event):
    pass

class IDAssigner(object):
    def __init__(self):
        self._indices = {}
        self._revindices = {}
        # Reserve 0 and 0xffff
        self._revindices[0] = '<reserve0>'
        self._revindices[0xffff] = '<reserve65535>'
        self._lastindex = 1
    def assign(self, key):
        if key in self._indices:
            return self._indices[key]
        else:
            ind = self._lastindex
            while ind in self._revindices:
                ind += 1
                ind &= 0xffff
            self._revindices[ind] = key
            self._indices[key] = ind
            self._lastindex = ind + 1
            return ind
    def unassign(self, keys):
        for k in keys:
            ind = self._indices.pop(k, None)
            if ind is not None:
                del self._revindices[ind]
    def frozen(self):
        return dict(self._indices)

def _to32bitport(portno):
    if portno >= 0xff00:
        portno = 0xffff0000 | portno
    return portno


@withIndices('connection')
class FlowReadyEvent(Event):
    pass


class IOFlowUpdater(FlowUpdater):
    def __init__(self, connection, systemid, bridgename, parent):
        FlowUpdater.__init__(self, connection, (PhysicalPortSet.default_key(),),
                                            ('ioprocessing', connection),
                                            parent._logger)
        self._walkerdict = {PhysicalPortSet.default_key(): partial(self._physicalport_walker, _portnames={})}
        self._systemid = systemid
        self._bridgename = bridgename
        self._portnames = {}
        self._portids = {}
        self._currentportids = {}
        self._currentportnames = {}
        self._lastportids = {}
        self._lastportnames = {}
        self._lastnetworkids = {}
        self._networkids = IDAssigner()
        self._phynetworkids = IDAssigner()
        self._physicalnetworkids = {}
        self._logicalportkeys = set()
        self._physicalportkeys = set()
        self._logicalnetworkkeys = set()
        self._physicalnetworkkeys = set()
        self._original_initialkeys = []
        self._append_initialkeys = []
        self._parent = parent
        self._flows_sent = set()

    async def update_ports(self, ports, ovsdb_ports):
        """
        Called from main module to update port information
        """
        new_port_names = dict((p['name'], _to32bitport(p['ofport'])) for p in ovsdb_ports)
        new_port_ids = dict((p['id'], _to32bitport(p['ofport'])) for p in ovsdb_ports if p['id'])
        if new_port_names == self._portnames and new_port_ids == self._portids:
            return
        self._portnames.clear()
        self._portnames.update(new_port_names)
        self._portids.clear()
        self._portids.update(new_port_ids)

        logicalportkeys = [LogicalPort.default_key(id) for id in self._portids]

        self._original_initialkeys = logicalportkeys + [PhysicalPortSet.default_key()]
        self._initialkeys = tuple(itertools.chain(self._original_initialkeys, self._append_initialkeys))
        phy_walker = partial(self._physicalport_walker, _portnames=new_port_names)
        log_walker = partial(self._logicalport_walker, _portids=new_port_ids)
        self._walkerdict = dict(itertools.chain(
            ((PhysicalPortSet.default_key(),phy_walker),),
            ((lgportkey,log_walker) for lgportkey in logicalportkeys)
        ))
        self._portnames = new_port_names
        self._portids = new_port_ids
        await self.restart_walk()

    async def flowready(self, logicalnetworkid, physicalportid):
        # 1. Check the current updated flows
        # 2. Check the current logicalnetwork and physicalport
        # 3. Wait for:
        #    a. flow updated event
        #    b. data object change event
        #    c. connection down event

        flowready_matcher = FlowReadyEvent.createMatcher(self._connection)
        conn_down = self._connection.protocol.statematcher(self._connection)
        dataobjectchanged = DataObjectChanged.createMatcher(None, None, self._connection)
        while self._connection.connected:
            currentlognetid = dict((id, n) for n, id in self._lastlognets)
            currentphyportid = dict((id, (p, p.physicalnetwork)) for p, id in self._lastphyports)
            if (logicalnetworkid, physicalportid) in self._flows_sent:
                return True
            elif logicalnetworkid in currentlognetid and physicalportid in currentphyportid:
                conn_down = OpenflowConnectionStateEvent.createMatcher(None, None, OpenflowConnectionStateEvent.CONNECTION_DOWN, self._connection)
                await M_(dataobjectchanged, conn_down, flowready_matcher)
            else:
                return False
        return False

    def _logicalport_walker(self, key, value, walk, save, _portids):
        _, (id,) = LogicalPort._getIndices(key)
        if id not in _portids:
            return
        save(key)
        if value is None:
            return
        with suppress(WalkKeyNotRetrieved):
            lognet = walk(value.network.getkey())
            save(lognet.getkey())
            phynet = walk(lognet.physicalnetwork.getkey())
            save(phynet.getkey())
        if hasattr(value,"subnet"):
            with suppress(WalkKeyNotRetrieved):
                subnet = walk(value.subnet.getkey())
                save(subnet.getkey())
                if hasattr(subnet,"router"):
                    routerport = walk(subnet.router.getkey())
                    save(routerport.getkey())
                    if hasattr(routerport,"router"):
                        router = walk(routerport.router.getkey())
                        save(router.getkey())
                        if router.interfaces.dataset():
                            for weakobj in router.interfaces.dataset():
                                with suppress(WalkKeyNotRetrieved):
                                    weakrouterport = walk(weakobj.getkey())
                                    save(weakrouterport.getkey())
                                    s = walk(weakrouterport.subnet.getkey())
                                    save(s.getkey())
                                    lgnet = walk(s.network.getkey())
                                    save(lgnet.getkey())
    def _physicalport_walker(self, key, value, walk, save, _portnames):
        save(key)
        if value is None:
            return
        physet = value.set
        for name in _portnames:
            phyports = physet.find(PhysicalPort, self._connection.protocol.vhost, self._systemid, self._bridgename, name)
            # There might be more than one match physical port rule for one port, pick the most specified one
            namedict = {}
            for p in phyports:
                _, inds = PhysicalPort._getIndices(p.getkey())
                name = inds[-1]
                ind_key = [i != '%' for i in inds]
                if name != '%':
                    if name in namedict:
                        if namedict[name][0] < ind_key:
                            namedict[name] = (ind_key, p)
                    else:
                        namedict[name] = (ind_key, p)
            phyports = [v[1] for v in namedict.values()]
            for p in phyports:
                with suppress(WalkKeyNotRetrieved):
                    phyp = walk(p.getkey())
                    save(phyp.getkey())
                    phynet = walk(phyp.physicalnetwork.getkey())
                    save(phynet.getkey())
                    if self._parent.enable_router_forward:
                        phynetmap = walk(PhysicalNetworkMap.default_key(phynet.id))
                        save(phynetmap.getkey())
                        for weak_lgnet in  phynetmap.logicnetworks.dataset():
                            with suppress(WalkKeyNotRetrieved):
                                lgnet = walk(weak_lgnet.getkey())
                                save(lgnet.getkey())

    def reset_initialkeys(self,keys,values):

        subnetkeys = [k for k,v in zip(keys,values) if v is not None and not v.isdeleted() and
                      v.isinstance(SubNet)]
        routerportkeys = [k for k,v in zip(keys,values) if v is not None and not v.isdeleted() and
                          v.isinstance(RouterPort)]
        portkeys = [k for k,v in zip(keys,values) if v is not None and not v.isdeleted() and
                    v.isinstance(VRouter)]
        self._append_initialkeys = subnetkeys + routerportkeys + portkeys
        self._initialkeys = tuple(itertools.chain(self._original_initialkeys, self._append_initialkeys))

    async def walkcomplete(self, keys, values):
        conn = self._connection
        dpid = conn.openflow_datapathid
        vhost = conn.protocol.vhost
        _currentportids = dict(self._portids)
        _currentportnames = dict(self._portnames)
        updated_data = {}
        current_data = {}
        for cls, name, idg, assigner in ((LogicalPort, '_logicalportkeys', lambda x: _currentportids.get(x.id), None),
                                 (PhysicalPort, '_physicalportkeys', lambda x: _currentportnames.get(x.name), None),
                                 (LogicalNetwork, '_logicalnetworkkeys', lambda x: self._networkids.assign(x.getkey()), self._networkids),
                                 (PhysicalNetwork, '_physicalnetworkkeys', lambda x: self._phynetworkids.assign(x.getkey()), self._phynetworkids),
                                 ):
            objs = [v for v in values if v is not None and not v.isdeleted() and v.isinstance(cls)]
            cv = [(o, oid) for o,oid in ((o, idg(o)) for o in objs) if oid is not None]
            objkeys = set([v.getkey() for v,_ in cv])
            oldkeys = getattr(self, name)
            current_data[cls] = cv
            if objkeys != oldkeys:
                if assigner is not None:
                    assigner.unassign(oldkeys.difference(objkeys))
                setattr(self, name, objkeys)
                updated_data[cls] = True
        if updated_data:
            await self.wait_for_send(DataObjectChanged(dpid, vhost, conn, LogicalPort in updated_data,
                                                                            PhysicalPort in updated_data,
                                                                            LogicalNetwork in updated_data,
                                                                            PhysicalNetwork in updated_data,
                                                                            current = (current_data.get(LogicalPort),
                                                                                       current_data.get(PhysicalPort),
                                                                                       current_data.get(LogicalNetwork),
                                                                                       current_data.get(PhysicalNetwork))))
        self._lastlognets = current_data.get(LogicalNetwork)
        self._lastphyports = current_data.get(PhysicalPort)
        self._currentportids = _currentportids
        self._currentportnames = _currentportnames

    async def updateflow(self, connection, addvalues, removevalues, updatedvalues):
        # We must do these in order, each with a batch:
        # 1. Remove flows
        # 2. Remove groups
        # 3. Add groups, modify groups
        # 4. Add flows, modify flows
        try:
            cmds = []
            ofdef = connection.openflowdef
            vhost = connection.protocol.vhost
            input_table = self._parent._gettableindex('ingress', vhost)
            input_next = self._parent._getnexttable('', 'ingress', vhost = vhost)
            output_table = self._parent._gettableindex('egress', vhost)
            # Cache all IDs, save them into last. We will need them for remove.
            _lastportids = self._lastportids
            _lastportnames = self._lastportnames
            _lastnetworkids = self._lastnetworkids
            _portids = dict(self._currentportids)
            _portnames = dict(self._currentportnames)
            _networkids = self._networkids.frozen()
            exist_objs = dict((obj.getkey(), obj) for obj in self._savedresult if obj is not None and not obj.isdeleted())
            # We must generate actions from network driver
            phyportset = [obj for obj in self._savedresult if obj is not None and not obj.isdeleted() and obj.isinstance(PhysicalPort)]
            phynetset = [obj for obj in self._savedresult if obj is not None and not obj.isdeleted() and obj.isinstance(PhysicalNetwork)]
            lognetset = [obj for obj in self._savedresult if obj is not None and not obj.isdeleted() and obj.isinstance(LogicalNetwork)]
            logportset = [obj for obj in self._savedresult if obj is not None and not obj.isdeleted() and obj.isinstance(LogicalPort)]
            # If a port is both a logical port and a physical port, flows may conflict.
            # Remove the port from dictionary if it is duplicated.
            logportofps = set(_portids[lp.id] for lp in logportset if lp.id in _portids)
            _portnames = dict((n,v) for n,v in _portnames.items() if v not in logportofps)
            self._lastportids = _portids
            self._lastportnames = _portnames
            self._lastnetworkids = _networkids
            # Group current ports by network for further use
            phyportdict = {}
            for p in phyportset:
                phyportdict.setdefault(p.physicalnetwork, []).append(p)
            lognetdict = {}
            for n in lognetset:
                lognetdict.setdefault(n.physicalnetwork, []).append(n)
            logportdict = {}
            for p in logportset:
                logportdict.setdefault(p.network, []).append(p)
            allapis = []
            # Updated networks when:
            # 1. Network is updated
            # 2. Physical network of this logical network is updated
            # 3. Logical port is added or removed from the network
            # 4. Physical port is added or removed from the physical network
            group_updates = set([obj for obj in updatedvalues if obj.isinstance(LogicalNetwork)])
            group_updates.update(obj.network for obj in addvalues if obj.isinstance(LogicalPort))
            #group_updates.update(obj.network for obj in updatedvalues if obj.isinstance(LogicalPort))
            group_updates.update(exist_objs[obj.network.getkey()] for obj in removevalues if obj.isinstance(LogicalPort) and obj.network.getkey() in exist_objs)
            updated_physicalnetworks = set(obj for obj in updatedvalues if obj.isinstance(PhysicalNetwork))
            updated_physicalnetworks.update(p.physicalnetwork for p in addvalues if p.isinstance(PhysicalPort))
            updated_physicalnetworks.update(exist_objs[p.physicalnetwork.getkey()] for p in removevalues if p.isinstance(PhysicalPort) and p.physicalnetwork.getkey() in exist_objs)
            updated_physicalnetworks.update(p.physicalnetwork for p in updatedvalues if p.isinstance(PhysicalPort))
            group_updates.update(lnet for pnet in updated_physicalnetworks
                                 if pnet in lognetdict
                                 for lnet in lognetdict[pnet])
            _flows_sent = set()
            for pnet in phynetset:
                if pnet in lognetdict and pnet in phyportdict:
                    for lognet in lognetdict[pnet]:
                        netid = _networkids.get(lognet.getkey())
                        if netid is not None:
                            for p in phyportdict[pnet]:
                                if lognet in addvalues or lognet in group_updates or p in addvalues or p in updatedvalues:
                                    pid = _portnames.get(p.name)
                                    if pid is not None:
                                        async def subr(lognet, p, netid, pid):
                                            try:
                                                r = await call_api(self, 'public', 'createioflowparts', {'connection': connection,
                                                                                                       'logicalnetwork': lognet,
                                                                                                       'physicalport': p,
                                                                                                       'logicalnetworkid': netid,
                                                                                                       'physicalportid': pid})
                                            except Exception:
                                                self._parent._logger.warning("Create flow parts failed for %r and %r", lognet, p, exc_info = True)
                                                return None
                                            else:
                                                _flows_sent.add((netid, pid))
                                                return ((lognet, p), r)
                                        allapis.append(subr(lognet, p, netid, pid))
            flowparts_result = await self.execute_all(allapis)
            flowparts = dict(r for r in flowparts_result if r is not None)
            if connection.protocol.disablenxext:
                # Nicira extension is disabled, use metadata instead
                # 64-bit metadata is used as:
                # | 16-bit input network | 16-bit output network | 16-bit reserved | 16-bit output port |
                # When first initialized, input network = output network = Logical Network no.
                # output port = OFPP_ANY, reserved bits are 0x0000
                # Currently used reserved bits:
                # left-most (offset = 15, mask = 0x8000): allow output to IN_PORT
                # offset = 14, mask = 0x4000: 1 if is IN_PORT is a logical port, 0 else
                # right-most (offset = 0, mask = 0x0001): VXLAN learned
                def create_input_instructions(lognetid, extra_actions, is_logport):
                    lognetid = (lognetid & 0xffff)
                    instructions = [ofdef.ofp_instruction_write_metadata(
                                        metadata = (lognetid << 48) | (lognetid << 32) | ((0x4000 if is_logport else 0) << 16) | (ofdef.OFPP_ANY & 0xffff),
                                        metadata_mask = 0xffffffffffffffff
                                    ),
                                    ofdef.ofp_instruction_goto_table(table_id = input_next)
                                    ]
                    if extra_actions:
                        instructions.insert(0, ofdef.ofp_instruction_actions(actions = list(extra_actions)))
                    return instructions
                def create_output_oxm(lognetid, portid, in_port = False):
                    r = [ofdef.create_oxm(ofdef.OXM_OF_METADATA_W, (portid & 0xFFFF) | (0x80000000 if in_port else 0) | ((lognetid & 0xFFFF) << 32), 0x0000FFFF8000FFFF)]
                    if in_port:
                        r.append(ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, portid))
                    return r
            else:
                # With nicira extension, we store input network, output network and output port in REG4, REG5 and REG6
                # REG7 is used as the reserved bits
                def create_input_instructions(lognetid, extra_actions, is_logport):
                    lognetid = (lognetid & 0xffff)
                    return [ofdef.ofp_instruction_actions(actions = [
                                    ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(ofdef.NXM_NX_REG4, lognetid)
                                            ),
                                    ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(ofdef.NXM_NX_REG5, lognetid)
                                            ),
                                    ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(ofdef.NXM_NX_REG6, ofdef.OFPP_ANY)
                                            ),
                                    ofdef.ofp_action_set_field(
                                            field = ofdef.create_oxm(ofdef.NXM_NX_REG7, (0x4000 if is_logport else 0))
                                            )
                                ] + list(extra_actions)),
                            ofdef.ofp_instruction_goto_table(table_id = input_next)
                            ]
                def create_output_oxm(lognetid, portid, in_port = False):
                    r = [ofdef.create_oxm(ofdef.NXM_NX_REG5, lognetid),
                            ofdef.create_oxm(ofdef.NXM_NX_REG6, portid),
                            ofdef.create_oxm(ofdef.NXM_NX_REG7_W, 0x8000 if in_port else 0, 0x8000)]
                    if in_port:
                        r.append(ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, portid))
                    return r
            for obj in removevalues:
                if obj.isinstance(LogicalPort):
                    ofport = _lastportids.get(obj.id)
                    if ofport is not None:
                        cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                 ofport
                                                                                 )])
                                                       ))
                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofport,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm()))
                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_IN_PORT,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                    ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, ofport)])))
                elif obj.isinstance(PhysicalPort):
                    ofport = _lastportnames.get(obj.name)
                    if ofport is not None:
                        cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                 ofport
                                                                                 )])
                                                       ))
                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                       cookie = 0x0001000000000000 | ((ofport & 0xffff) << 16),
                                                       cookie_mask = 0xffffffffffff0000,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm()))
                elif obj.isinstance(LogicalNetwork):
                    groupid = _lastnetworkids.get(obj.getkey())
                    if groupid is not None:
                        cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                       cookie = 0x0001000000000000 | groupid,
                                                       cookie_mask = 0xffffffffffffffff,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm()
                                                       ))
                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                       cookie = 0x0001000000000000 | groupid,
                                                       cookie_mask = 0xffff00000000ffff,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm()
                                                       ))
                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(groupid, ofdef.OFPP_ANY))
                                                       ))
            # Never use flow mod to update an input flow of physical port, because the input_oxm may change.
            for obj in updatedvalues:
                if obj.isinstance(PhysicalPort):
                    ofport = _portnames.get(obj.name)
                    if ofport is not None:
                        cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                 ofport
                                                                                 )])
                                                       ))
                elif obj.isinstance(LogicalNetwork):
                    groupid = _networkids.get(obj.getkey())
                    if groupid is not None:
                        cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                       cookie = 0x0001000000000000 | groupid,
                                                       cookie_mask = 0xffffffffffffffff,
                                                       command = ofdef.OFPFC_DELETE,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm()
                                                       ))
                elif obj.isinstance(PhysicalNetwork):
                    if obj in phyportdict:
                        for p in phyportdict[obj]:
                            ofport = _portnames.get(p.name)
                            if ofport is not None and p not in addvalues and p not in updatedvalues:
                                cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                               command = ofdef.OFPFC_DELETE,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                        ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                         ofport
                                                                                         )])
                                                               ))
            await self.execute_commands(connection, cmds)
            del cmds[:]
            for obj in removevalues:
                if obj.isinstance(LogicalNetwork):
                    groupid = _lastnetworkids.get(obj.getkey())
                    if groupid is not None:
                        cmds.append(ofdef.ofp_group_mod(command = ofdef.OFPGC_DELETE,
                                                        type = ofdef.OFPGT_ALL,
                                                        group_id = groupid
                                                        ))
            await self.execute_commands(connection, cmds)
            del cmds[:]
            disablechaining = connection.protocol.disablechaining
            created_groups = {}
            def create_buckets(obj, groupid):
                # Generate buckets
                buckets = [ofdef.ofp_bucket(actions=[ofdef.ofp_action_output(port = _portids[p.id])])
                           for p in logportdict[obj]
                           if p.id in _portids] if obj in logportdict else []
                allactions = [ofdef.ofp_action_output(port = _portids[p.id])
                              for p in logportdict[obj]
                              if p.id in _portids] if obj in logportdict else []
                disablegroup = False
                if obj.physicalnetwork in phyportdict:
                    for p in phyportdict[obj.physicalnetwork]:
                        if (obj, p) in flowparts:
                            fp = flowparts[(obj,p)]
                            allactions.extend(fp[3])
                            if disablechaining and not disablegroup and any(a.type == ofdef.OFPAT_GROUP for a in fp[3]):
                                # We cannot use chaining. We use a long action list instead, and hope there is no conflicts
                                disablegroup = True
                            else:
                                buckets.append(ofdef.ofp_bucket(actions=list(fp[3])))
                if disablegroup:
                    created_groups[groupid] = allactions
                else:
                    created_groups[groupid] = [ofdef.ofp_action_group(group_id = groupid)]
                return buckets
            for obj in addvalues:
                if obj.isinstance(LogicalNetwork):
                    groupid = _networkids.get(obj.getkey())
                    if groupid is not None:
                        cmds.append(ofdef.ofp_group_mod(command = ofdef.OFPGC_ADD,
                                                        type = ofdef.OFPGT_ALL,
                                                        group_id = groupid,
                                                        buckets = create_buckets(obj, groupid)
                                                        ))
            for obj in group_updates:
                groupid = _networkids.get(obj.getkey())
                if groupid is not None:
                    cmds.append(ofdef.ofp_group_mod(command = ofdef.OFPGC_MODIFY,
                                                    type = ofdef.OFPGT_ALL,
                                                    group_id = groupid,
                                                    buckets = create_buckets(obj, groupid)
                                                    ))
            await self.execute_commands(connection, cmds)
            del cmds[:]
            # There are 5 kinds of flows:
            # 1. in_port = (Logical Port)
            # 2. in_port = (Physical_Port), network = (Logical_Network)
            # 3. out_port = (Logical Port)
            # 4. out_port = (Physical_Port), network = (Logical_Network)
            # 5. out_port = OFPP_ANY, network = (Logical_Network)
            for obj in addvalues:
                if obj.isinstance(LogicalPort):
                    ofport = _portids.get(obj.id)
                    lognetid = _networkids.get(obj.network.getkey())
                    if ofport is not None and lognetid is not None:
                        cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                       command = ofdef.OFPFC_ADD,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                 ofport
                                                                                 )]),
                                                       instructions = create_input_instructions(lognetid, [], True)
                                                       ))
                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                       command = ofdef.OFPFC_ADD,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport)),
                                                       instructions = [ofdef.ofp_instruction_actions(actions = [
                                                                    ofdef.ofp_action_output(port = ofport)
                                                                    ])]
                                                       ))
                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                       command = ofdef.OFPFC_ADD,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport, True)),
                                                       instructions = [ofdef.ofp_instruction_actions(actions = [
                                                                    ofdef.ofp_action_output(port = ofdef.OFPP_IN_PORT)
                                                                    ])]
                                                       ))
            # Ignore update of logical port
            # Physical port:
            for obj in addvalues:
                if obj.isinstance(PhysicalPort):
                    ofport = _portnames.get(obj.name)
                    if ofport is not None and obj.physicalnetwork in lognetdict:
                        for lognet in lognetdict[obj.physicalnetwork]:
                            lognetid = _networkids.get(lognet.getkey())
                            if lognetid is not None and (lognet, obj) in flowparts:
                                input_oxm, input_actions, output_actions, _, output_actions2 = flowparts[(lognet, obj)]
                                cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                               cookie = 0x0001000000000000 | lognetid,
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                        ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                         ofport
                                                                                         )] + list(input_oxm)),
                                                               instructions = create_input_instructions(lognetid, input_actions, False)
                                                               ))
                                cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                               cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport, False)),
                                                               instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                            list(output_actions))]
                                                               ))
                                cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                               cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport, True)),
                                                               instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                            list(output_actions2))]
                                                               ))
            for lognet in addvalues:
                if lognet.isinstance(LogicalNetwork):
                    lognetid = _networkids.get(lognet.getkey())
                    if lognetid is not None and lognet.physicalnetwork in phyportdict:
                        for obj in phyportdict[lognet.physicalnetwork]:
                            ofport = _portnames.get(obj.name)
                            if ofport is not None and (lognet, obj) in flowparts and obj not in addvalues:
                                input_oxm, input_actions, output_actions, _, output_actions2 = flowparts[(lognet, obj)]
                                cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                               cookie = 0x0001000000000000 | lognetid,
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                        ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                         ofport
                                                                                         )] + input_oxm),
                                                               instructions = create_input_instructions(lognetid, input_actions, False)
                                                               ))
                                cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                               cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport, False)),
                                                               instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                            list(output_actions))]
                                                               ))
                                cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                               cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport, True)),
                                                               instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                            list(output_actions2))]
                                                               ))
            for obj in updatedvalues:
                if obj.isinstance(PhysicalPort):
                    ofport = _portnames.get(obj.name)
                    if ofport is not None and obj.physicalnetwork in lognetdict:
                        for lognet in lognetdict[obj.physicalnetwork]:
                            lognetid = _networkids.get(lognet.getkey())
                            if lognetid is not None and (lognet, obj) in flowparts and not lognet in addvalues:
                                input_oxm, input_actions, output_actions, _, output_actions2 = flowparts[(lognet, obj)]
                                cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                               cookie = 0x0001000000000000 | lognetid,
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                        ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                         ofport
                                                                                         )] + input_oxm),
                                                               instructions = create_input_instructions(lognetid, input_actions, False)
                                                               ))
                                cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                               cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_MODIFY,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport, False)),
                                                               instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                            list(output_actions))]
                                                               ))
                                cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                               cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_MODIFY,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport, True)),
                                                               instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                            list(output_actions2))]
                                                               ))
            for lognet in updatedvalues:
                if lognet.isinstance(LogicalNetwork):
                    lognetid = _networkids.get(lognet.getkey())
                    if lognetid is not None and lognet.physicalnetwork in phyportdict:
                        for obj in phyportdict[lognet.physicalnetwork]:
                            ofport = _portnames.get(obj.name)
                            if ofport is not None and (lognet, obj) in flowparts and obj not in addvalues and obj not in updatedvalues:
                                input_oxm, input_actions, output_actions, _, output_actions2 = flowparts[(lognet, obj)]
                                cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                               cookie = 0x0001000000000000 | lognetid,
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_ADD,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                        ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                         ofport
                                                                                         )] + input_oxm),
                                                               instructions = create_input_instructions(lognetid, input_actions, False)
                                                               ))
                                cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                               cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_MODIFY,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport)),
                                                               instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                            list(output_actions))]
                                                               ))
                                cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                               cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                               cookie_mask = 0xffffffffffffffff,
                                                               command = ofdef.OFPFC_MODIFY,
                                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                                               out_port = ofdef.OFPP_ANY,
                                                               out_group = ofdef.OFPG_ANY,
                                                               match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport, True)),
                                                               instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                            list(output_actions2))]
                                                               ))
            # Physical network is updated
            for pnet in updatedvalues:
                if pnet.isinstance(PhysicalNetwork) and pnet in lognetdict:
                    for lognet in lognetdict[pnet]:
                        if lognet.isinstance(LogicalNetwork):
                            lognetid = _networkids.get(lognet.getkey())
                            if lognetid is not None and lognet not in updatedvalues and lognet not in addvalues and lognet.physicalnetwork in phyportdict:
                                for obj in phyportdict[lognet.physicalnetwork]:
                                    ofport = _portnames.get(obj.name)
                                    if ofport is not None and (lognet, obj) in flowparts and obj not in addvalues and obj not in updatedvalues:
                                        input_oxm, input_actions, output_actions, _, output_actions2 = flowparts[(lognet, obj)]
                                        cmds.append(ofdef.ofp_flow_mod(table_id = input_table,
                                                                       cookie = 0x0001000000000000 | lognetid,
                                                                       cookie_mask = 0xffffffffffffffff,
                                                                       command = ofdef.OFPFC_ADD,
                                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                                       out_port = ofdef.OFPP_ANY,
                                                                       out_group = ofdef.OFPG_ANY,
                                                                       match = ofdef.ofp_match_oxm(oxm_fields = [
                                                                                ofdef.create_oxm(ofdef.OXM_OF_IN_PORT,
                                                                                                 ofport
                                                                                                 )] + input_oxm),
                                                                       instructions = create_input_instructions(lognetid, input_actions, False)
                                                                       ))
                                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                                       cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                                       cookie_mask = 0xffffffffffffffff,
                                                                       command = ofdef.OFPFC_MODIFY,
                                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                                       out_port = ofdef.OFPP_ANY,
                                                                       out_group = ofdef.OFPG_ANY,
                                                                       match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport)),
                                                                       instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                                    list(output_actions))]
                                                                       ))
                                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                                       cookie = 0x0001000000000000 | lognetid | ((ofport & 0xffff) << 16),
                                                                       cookie_mask = 0xffffffffffffffff,
                                                                       command = ofdef.OFPFC_MODIFY,
                                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                                       out_port = ofdef.OFPP_ANY,
                                                                       out_group = ofdef.OFPG_ANY,
                                                                       match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofport, True)),
                                                                       instructions = [ofdef.ofp_instruction_actions(actions = 
                                                                                    list(output_actions2))]
                                                                       ))
            # Logical network broadcast
            for lognet in addvalues:
                if lognet.isinstance(LogicalNetwork):
                    lognetid = _networkids.get(lognet.getkey())
                    if lognetid is not None and lognetid in created_groups:
                        cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                                       command = ofdef.OFPFC_ADD,
                                                       priority = ofdef.OFP_DEFAULT_PRIORITY,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofdef.OFPP_ANY)),
                                                       instructions = [ofdef.ofp_instruction_actions(actions =
                                                                            created_groups.pop(lognetid))]
                                                       ))
            for lognetid, actions in created_groups.items():
                cmds.append(ofdef.ofp_flow_mod(table_id = output_table,
                                               command = ofdef.OFPFC_ADD,
                                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                                               buffer_id = ofdef.OFP_NO_BUFFER,
                                               out_port = ofdef.OFPP_ANY,
                                               out_group = ofdef.OFPG_ANY,
                                               match = ofdef.ofp_match_oxm(oxm_fields = create_output_oxm(lognetid, ofdef.OFPP_ANY)),
                                               instructions = [ofdef.ofp_instruction_actions(actions = actions)]
                                               ))                
            # Ignore logical network update
            await self.execute_commands(connection, cmds)
            self._flows_sent = _flows_sent
            await self.wait_for_send(FlowReadyEvent(self._connection))

        except Exception:
            self._parent._logger.warning("Update flow for connection %r failed with exception", connection, exc_info = True)
            # We don't want the whole flow update stops, so ignore the exception and continue
    
@defaultconfig
@depend(ofpportmanager.OpenflowPortManager, ovsdbportmanager.OVSDBPortManager, objectdb.ObjectDB)
class IOProcessing(FlowBase):
    "Ingress and Egress processing"
    _tablerequest = (("ingress", (), ''),
                     ("egress", ("ingress",),''))
    # vHost map from OpenFlow vHost to OVSDB vHost. If the OpenFlow vHost is not found in this map,
    # it will map to the default OVSDB vHost ('')
    _default_vhostmap = {}
    # Enable forwarding in this server, so it becomes a forwarding node (also known as a N/S gateway)
    _default_enable_router_forward = False

    def __init__(self, server):
        FlowBase.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._main
        self.routines.append(self.apiroutine)
        self._flowupdaters = {}
        self._portchanging = set()
        self._portchanged = set()
        self.createAPI(api(self.flowready, self.apiroutine))

    async def flowready(self, connection, logicalnetworkid, physicalportid):
        """
        Wait until flows are sent to switch

        :param connection: Openflow connection
        :param logicalnetworkid: logical network id (integer)
        :param physicalportid: physical port id (integer)
        :return: If connection/network/port not exists, return False, else return True
        """
        if connection not in self._flowupdaters:
            return False
        else:
            return await self._flowupdaters[connection].flowready(logicalnetworkid, physicalportid)

    async def _main(self):
        flow_init = FlowInitialize.createMatcher(_ismatch = lambda x: self.vhostbind is None or x.vhost in self.vhostbind)
        port_change = ModuleNotification.createMatcher("openflowportmanager", "update", _ismatch = lambda x: self.vhostbind is None or x.vhost in self.vhostbind)
        while True:
            e, m = await M_(flow_init, port_change)
            c = e.connection
            if m is flow_init:
                self.apiroutine.subroutine(self._init_conn(c))
            else:
                if e.reason == 'disconnected':
                    self.apiroutine.subroutine(self._remove_conn(c))
                else:
                    self.apiroutine.subroutine(self._portchange(c))

    async def _init_conn(self, conn):
        # Default drop
        await conn.protocol.batch((conn.openflowdef.ofp_flow_mod(table_id = self._gettableindex("ingress", conn.protocol.vhost),
                                                           command = conn.openflowdef.OFPFC_ADD,
                                                           priority = 0,
                                                           buffer_id = conn.openflowdef.OFP_NO_BUFFER,
                                                           match = conn.openflowdef.ofp_match_oxm(),
                                                           instructions = [conn.openflowdef.ofp_instruction_actions(
                                                                            type = conn.openflowdef.OFPIT_CLEAR_ACTIONS
                                                                            )]
                                                           ),
                                      conn.openflowdef.ofp_flow_mod(table_id = self._gettableindex("egress", conn.protocol.vhost),
                                                           command = conn.openflowdef.OFPFC_ADD,
                                                           priority = 0,
                                                           buffer_id = conn.openflowdef.OFP_NO_BUFFER,
                                                           match = conn.openflowdef.ofp_match_oxm(),
                                                           instructions = [conn.openflowdef.ofp_instruction_actions(
                                                                            type = conn.openflowdef.OFPIT_CLEAR_ACTIONS
                                                                            )]
                                                           )), conn, self.apiroutine)
        if conn in self._flowupdaters:
            self._flowupdaters[conn].close()
        datapath_id = conn.openflow_datapathid
        ovsdb_vhost = self.vhostmap.get(conn.protocol.vhost, "")
        bridgename, systemid, _ = await call_api(self.apiroutine, 'ovsdbmanager', 'waitbridgeinfo',
                                                 {'datapathid': datapath_id,
                                                  'vhost': ovsdb_vhost})
        new_updater = IOFlowUpdater(conn, systemid, bridgename, self)
        self._flowupdaters[conn] = new_updater
        new_updater.start()
        await self._portchange(conn)

    async def _remove_conn(self, conn):
        # Do not need to modify flows
        if conn in self._flowupdaters:
            self._flowupdaters[conn].close()
            del self._flowupdaters[conn]

    async def _portchange(self, conn):
        # Do not re-enter
        if conn in self._portchanging:
            self._portchanged.add(conn)
            return
        self._portchanging.add(conn)
        last_portno = set()
        try:
            while True:
                self._portchanged.discard(conn)
                flow_updater = self._flowupdaters.get(conn)
                if flow_updater is None:
                    break
                if not conn.connected:
                    break
                datapath_id = conn.openflow_datapathid
                ovsdb_vhost = self.vhostmap.get(conn.protocol.vhost, "")
                ovsdb_update_event_matcher = ModuleNotification.createMatcher(
                                                "ovsdbportmanager",
                                                "update",
                                                _ismatch = lambda x: x.vhost == ovsdb_vhost and x.datapathid == datapath_id)
                ovsdb_updated = False
                def _ovsdb_update_callback(event, matcher):
                    nonlocal ovsdb_updated
                    ovsdb_updated = True
                ports, ovsdb_ports = \
                    await self.apiroutine.with_callback( 
                                    self.apiroutine.execute_all(
                                                    [call_api(self.apiroutine, 'openflowportmanager', 'getports', {'datapathid': datapath_id,
                                                                                            'vhost': conn.protocol.vhost}),
                                                     call_api(self.apiroutine, 'ovsdbportmanager', 'getports', {'datapathid': datapath_id,
                                                                                                  'vhost': ovsdb_vhost})]),
                                    _ovsdb_update_callback,
                                    ovsdb_update_event_matcher
                                )
                if conn in self._portchanged or ovsdb_updated:
                    # Retrieve again
                    continue
                if not conn.connected:
                    self._portchanged.discard(conn)
                    return
                ovsdb_port_dict = {p['ofport']: p for p in ovsdb_ports}
                # Choose the intersection of ports from two sources
                port_pairs = [(p, ovsdb_port_dict[p.port_no & 0xffff])
                              for p in ports
                              if (p.port_no & 0xffff) in ovsdb_port_dict]
                current_portno = {p.port_no for p, _ in port_pairs}
                # Get again to prevent concurrent problems
                flow_updater = self._flowupdaters.get(conn)
                if flow_updater is None:
                    break
                if not conn.connected:
                    break
                if conn in self._portchanged or ovsdb_updated:
                    continue
                # If all openflow ports have their OVSDB ports, we are in sync and can exit
                if all((p.port_no & 0xffff) in ovsdb_port_dict for p in ports):
                    if current_portno != last_portno:
                        if port_pairs:
                            await self.apiroutine.with_callback(
                                        flow_updater.update_ports(*zip(*port_pairs)),
                                        _ovsdb_update_callback,
                                        ovsdb_update_event_matcher
                                    )
                        else:
                            await self.apiroutine.with_callback(
                                        flow_updater.update_ports((), ()),
                                        _ovsdb_update_callback,
                                        ovsdb_update_event_matcher
                                    )
                    break
                else:
                    # Partially update
                    if current_portno and current_portno != last_portno:
                        if port_pairs:
                            await self.apiroutine.with_callback(
                                        flow_updater.update_ports(*zip(*port_pairs)),
                                        _ovsdb_update_callback,
                                        ovsdb_update_event_matcher
                                    )
                        else:
                            await self.apiroutine.with_callback(
                                        flow_updater.update_ports((), ()),
                                        _ovsdb_update_callback,
                                        ovsdb_update_event_matcher
                                    )
                        last_portno = current_portno                    
                    # Some openflow ports do not have OVSDB information, this may be caused
                    # by:
                    # 1. A port is added to OpenFlow, but not yet retrieved from OVSDB
                    # 2. A port is deleted from OVSDB, but not yet updated in OpenFlow
                    # 3. Other synchronization problem
                    port_change = ModuleNotification.createMatcher("openflowportmanager", "update",
                                                                   _ismatch = lambda x: x.connection == conn)
                    conndown = conn.protocol.statematcher(conn)
                    timeout, _, m = await self.apiroutine.wait_with_timeout(5,
                                                                            port_change,
                                                                            ovsdb_update_event_matcher,
                                                                            conndown)
                    if timeout:
                        self._logger.warning('OpenFlow ports may not be synchronized. Try resync...')
                        # Connection is up but ports are not synchronized, try resync
                        await self.apiroutine.execute_all([call_api(self.apiroutine, 'openflowportmanager', 'resync',
                                                                     {'datapathid': datapath_id,
                                                                      'vhost': conn.protocol.vhost}),
                                                           call_api(self.apiroutine, 'ovsdbportmanager', 'resync',
                                                                     {'datapathid': datapath_id,
                                                                      'vhost': ovsdb_vhost})])
                        # Wait for a while
                        await self.apiroutine.wait_with_timeout(5)
                        continue
                    elif m is conndown:
                        # Connection lost, no longer need to trace the port changes
                        break
        finally:
            self._portchanging.remove(conn)



'''
Created on 2016/7/4

:author: hubo
'''


from vlcp.config.config import defaultconfig
from vlcp.server.module import depend, api
from vlcp.service.sdn.flowbase import FlowBase
from vlcp.service.sdn.ofpmanager import FlowInitialize
from vlcp.utils.flowupdater import FlowUpdater
from vlcp.utils.networkmodel import PhysicalPort, LogicalPort, LogicalNetwork,\
    LogicalNetworkMap
import vlcp.service.kvdb.objectdb as objectdb
import vlcp.service.sdn.ofpportmanager as ofpportmanager
from vlcp.event.runnable import RoutineContainer
from vlcp.protocol.openflow.openflow import OpenflowConnectionStateEvent,\
    OpenflowAsyncMessageEvent, OpenflowErrorResultException
from vlcp.utils.ethernet import ethernet_l2
import vlcp.service.sdn.ioprocessing as iop
import itertools
from vlcp.utils.dataobject import ReferenceObject
from vlcp.utils.exceptions import WalkKeyNotRetrieved
from contextlib import suppress
from vlcp.event.event import M_


class ARPUpdater(FlowUpdater):
    def __init__(self, connection, parent):
        FlowUpdater.__init__(self, connection, (), ('ARPUpdater', connection), parent._logger)
        self._parent = parent
        self._lastlognets = ()
        self._lastphyports = ()
        self._lastlogports = ()
        self._lastlogportinfo = {}
        self._lastphyportinfo = {}
        self._lastlognetinfo = {}
        self._last_arps = set()
        self._last_freearps = {}

    async def main(self):
        try:
            if self._connection.protocol.disablenxext:
                self._logger.warning("ARP responder disabled on connection %r because Nicira extension is not enabled", self._connection)
                return
            self.subroutine(self._update_handler(), True, '_update_handler_routine')
            await FlowUpdater.main(self)
        finally:
            if hasattr(self, '_update_handler_routine'):
                self._update_handler_routine.close()

    async def _update_handler(self):
        dataobjectchanged = iop.DataObjectChanged.createMatcher(None, None, self._connection)
        while True:
            ev = await dataobjectchanged
            self._lastlogports, self._lastphyports, self._lastlognets, _ = ev.current
            self._update_walk()
            self.updateobjects((p for p,_ in self._lastlogports))

    def _walk_logport(self, key, value, walk, save):
        if value is not None:
            save(key)

    def _walk_phyport(self, key, value, walk, save):
        if value is not None:
            save(key)

    def _walk_lognet(self, key, value, walk, save):
        save(key)
        if value is None:
            return
        if self._parent.prepush:
            # Acquire all logical ports
            with suppress(WalkKeyNotRetrieved):
                netmap = walk(LogicalNetworkMap.default_key(value.id))
                save(netmap.getkey())
                for logport in netmap.ports.dataset():
                    with suppress(WalkKeyNotRetrieved):
                        p = walk(logport.getkey())
                        #if p is not None and hasattr(p, 'mac_address') and hasattr(p, 'ip_address'):
                        save(logport.getkey())

    def _update_walk(self):
        logport_keys = [p.getkey() for p,_ in self._lastlogports]
        phyport_keys = [p.getkey() for p,_ in self._lastphyports]
        lognet_keys = [n.getkey() for n,_ in self._lastlognets]
        lognet_mapkeys = [LogicalNetworkMap.default_key(n.id) for n,_ in self._lastlognets]
        if self._parent.prepush:
            self._initialkeys = logport_keys + lognet_keys + phyport_keys + lognet_mapkeys
        else:
            self._initialkeys = logport_keys + lognet_keys + phyport_keys
        self._walkerdict = dict(itertools.chain(((n, self._walk_lognet) for n in lognet_keys),
                                                ((p, self._walk_logport) for p in logport_keys),
                                                ((p, self._walk_phyport) for p in phyport_keys)))
        self.subroutine(self.restart_walk(), False)

    async def updateflow(self, connection, addvalues, removevalues, updatedvalues):
        try:
            allobjs = set(o for o in self._savedresult if o is not None and not o.isdeleted())
            lastlogportinfo = self._lastlogportinfo
            lastlognetinfo = self._lastlognetinfo
            lastphyportinfo = self._lastphyportinfo
            lastfreearps = self._last_freearps
            currentlognetinfo = dict((n,(id, n.physicalnetwork)) for n,id in self._lastlognets if n in allobjs)
            netdict = dict((n.getkey(), n) for n in currentlognetinfo)
            currentlogportinfo = dict((p, (id, p.network)) for p,id in self._lastlogports if p in allobjs)
            currentphyportinfo = dict((p, (id, p.physicalnetwork)) for p,id in self._lastphyports if p in allobjs)
            last_arps = self._last_arps
            current_arps = {}
            if connection in self._parent._extra_arps:
                _arps_list = self._parent._extra_arps[connection]
                for ip, mac, lognetid, islocal in _arps_list:
                    lognet = netdict.get(LogicalNetwork.default_key(lognetid))
                    if lognet is not None:
                        if islocal is None:
                            arp_set = current_arps.setdefault(lognet, {})
                            arp_set[(ip, True)] = (mac, False)
                            arp_set[(ip, False)] = (mac, False)
                        else:
                            arp_set = current_arps.setdefault(lognet, {})
                            arp_set[(ip, islocal)] = (mac, False)
            broadcast_only = self._parent.broadcastonly
            currentfreearps = {}
            if self._parent.prepush:
                for obj in self._savedresult:
                    if obj is not None and not obj.isdeleted() and obj.isinstance(LogicalPort) and hasattr(obj, 'ip_address') and hasattr(obj, 'mac_address'):
                        if obj.network in currentlognetinfo:
                            arp_set = current_arps.setdefault(obj.network, {})
                            arp_set[(obj.ip_address, True)] = (obj.mac_address, broadcast_only)
                            if obj in currentlogportinfo:
                                # This port is in local switch, add a free ARP flow
                                pid, network = currentlogportinfo[obj]
                                currentfreearps[(network, obj.ip_address, obj)] = (obj.mac_address, broadcast_only,
                                                                                       network.physicalnetwork.type in self._parent.enable_freearp_networktypes and \
                                                                                        self._parent.enable_freearp)
            # Flow mapping:
            
            # arp_set => create_flow
            # currentfreearps => create_flow2
            self._lastlogportinfo = currentlogportinfo
            self._lastlognetinfo = currentlognetinfo
            self._lastphyportinfo = currentphyportinfo
            self._last_arps = current_arps
            self._last_freearps = currentfreearps
            cmds = []
            ofdef = connection.openflowdef
            vhost = connection.protocol.vhost
            arp = self._parent._gettableindex('arp', vhost)
            l2out_next = self._parent._getnexttable('', 'l2output', vhost = vhost)
            arp_next = self._parent._getnexttable('', 'arp', vhost = vhost)
            #===================================================================
            # if connection.protocol.disablenxext:
            #     def match_network(nid):
            #         return ofdef.create_oxm(ofdef.OXM_OF_METADATA_W, (nid & 0xFFFF) << 48, 0xFFFF000000000000)
            #     def create_instructions(actions, pid):
            #         return [ofdef.ofp_instruction_actions(actions = actions),
            #                 ofdef.ofp_instruction_write_metadata(metadata = 0x0000000080000000 | (pid & 0xffff),
            #                                                      metadata_mask = 0x000000008000ffff),
            #                 ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
            #===================================================================
            #else:
            def match_network(nid):
                return ofdef.create_oxm(ofdef.NXM_NX_REG4, nid)
            def create_instructions(actions):
                return [ofdef.ofp_instruction_actions(actions = actions + \
                                                      [ofdef.nx_action_reg_load(
                                                            ofs_nbits = ofdef.create_ofs_nbits(15,1),
                                                            dst = ofdef.NXM_NX_REG7,
                                                            value = 1
                                                        ),
                                                       ofdef.nx_action_reg_move(
                                                            n_bits = 32,
                                                            src = ofdef.OXM_OF_IN_PORT,
                                                            dst = ofdef.NXM_NX_REG6
                                                            )]),
                        ofdef.ofp_instruction_goto_table(table_id = l2out_next)]
            for n in last_arps:
                if n not in current_arps:
                    if n in lastlognetinfo:
                        nid, _ = lastlognetinfo[n]
                        for ip, islocal in last_arps[n]:
                            cmds.append(ofdef.ofp_flow_mod(table_id = arp,
                                                           cookie = 0x1 | (0x2 if islocal else 0),
                                                           cookie_mask = 0x7,
                                                           command = ofdef.OFPFC_DELETE,
                                                           buffer_id = ofdef.OFP_NO_BUFFER,
                                                           out_port = ofdef.OFPP_ANY,
                                                           out_group = ofdef.OFPG_ANY,
                                                           match = ofdef.ofp_match_oxm(
                                                                        oxm_fields = [match_network(nid),
                                                                                      ofdef.create_oxm(
                                                                                          ofdef.OXM_OF_ETH_TYPE,
                                                                                          ofdef.ETHERTYPE_ARP),
                                                                                      ofdef.create_oxm(
                                                                                          ofdef.OXM_OF_ARP_TPA,
                                                                                          ofdef.ip4_addr(ip)),
                                                                                      ofdef.create_oxm(
                                                                                          ofdef.OXM_OF_ARP_OP,
                                                                                          ofdef.ARPOP_REQUEST)
                                                                                      ]
                                                                    )
                                                           ))
                else:
                    if n in lastlognetinfo and n in currentlognetinfo and currentlognetinfo[n] != lastlognetinfo[n]:
                        # Delete all because network changes
                        nid, _ = lastlognetinfo[n]
                        for ip, islocal in last_arps[n]:
                            cmds.append(ofdef.ofp_flow_mod(table_id = arp,
                                                           cookie = 0x1 | (0x2 if islocal else 0),
                                                           cookie_mask = 0x7,
                                                           command = ofdef.OFPFC_DELETE,
                                                           buffer_id = ofdef.OFP_NO_BUFFER,
                                                           out_port = ofdef.OFPP_ANY,
                                                           out_group = ofdef.OFPG_ANY,
                                                           match = ofdef.ofp_match_oxm(
                                                                        oxm_fields = [match_network(nid),
                                                                                      ofdef.create_oxm(
                                                                                          ofdef.OXM_OF_ETH_TYPE,
                                                                                          ofdef.ETHERTYPE_ARP),
                                                                                      ofdef.create_oxm(
                                                                                          ofdef.OXM_OF_ARP_TPA,
                                                                                          ofdef.ip4_addr(ip)),
                                                                                      ofdef.create_oxm(
                                                                                          ofdef.OXM_OF_ARP_OP,
                                                                                          ofdef.ARPOP_REQUEST)
                                                                                      ]
                                                                    )
                                                           ))
                    else:
                        if n in currentlognetinfo:
                            nid, _ = currentlognetinfo[n]
                            for (ip,islocal), value in last_arps[n].items():
                                if (ip,islocal) not in current_arps[n] or value != current_arps[n]:
                                    cmds.append(ofdef.ofp_flow_mod(table_id = arp,
                                                                   cookie = 0x1 | (0x2 if islocal else 0),
                                                                   cookie_mask = 0x7,
                                                                   command = ofdef.OFPFC_DELETE,
                                                                   buffer_id = ofdef.OFP_NO_BUFFER,
                                                                   out_port = ofdef.OFPP_ANY,
                                                                   out_group = ofdef.OFPG_ANY,
                                                                   match = ofdef.ofp_match_oxm(
                                                                                oxm_fields = [match_network(nid),
                                                                                              ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE, ofdef.ETHERTYPE_ARP),
                                                                                              ofdef.create_oxm(ofdef.OXM_OF_ARP_TPA, ofdef.ip4_addr(ip)),
                                                                                              ofdef.create_oxm(ofdef.OXM_OF_ARP_OP, ofdef.ARPOP_REQUEST)
                                                                                              ]
                                                                            )
                                                                   ))
            for (network, ip, port), value in lastfreearps.items():
                if (network, ip, port) not in currentfreearps or currentfreearps[(network, ip, port)] != value \
                        or lastlogportinfo.get(port) != currentlogportinfo.get(port) \
                        or lastlognetinfo.get(network) != currentlognetinfo.get(network):
                    if port in lastlogportinfo and network in lastlognetinfo:
                        pid, _ = lastlogportinfo[port]
                        nid, _ = lastlognetinfo[network]
                        cmds.append(ofdef.ofp_flow_mod(table_id = arp,
                                                       cookie = 0x5 | (0x2 if islocal else 0),
                                                       cookie_mask = 0x7,
                                                       command = ofdef.OFPFC_DELETE,
                                                       buffer_id = ofdef.OFP_NO_BUFFER,
                                                       out_port = ofdef.OFPP_ANY,
                                                       out_group = ofdef.OFPG_ANY,
                                                       match = ofdef.ofp_match_oxm(
                                                                    oxm_fields = [ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, pid),
                                                                                  match_network(nid),
                                                                                  ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE, ofdef.ETHERTYPE_ARP),
                                                                                  ofdef.create_oxm(ofdef.OXM_OF_ARP_TPA, ofdef.ip4_addr(ip)),
                                                                                  ofdef.create_oxm(ofdef.OXM_OF_ARP_OP, ofdef.ARPOP_REQUEST)
                                                                                  ]
                                                                )
                                                       ))
            await self.execute_commands(connection, cmds)
            del cmds[:]
            # Create flows
            def _create_flow(ip, mac, nid, islocal, broadcast):
                return ofdef.ofp_flow_mod(
                               table_id = arp,
                               cookie = 0x1 | (0x2 if islocal else 0),
                               cookie_mask = 0xffffffffffffffff,
                               command = ofdef.OFPFC_ADD,
                               buffer_id = ofdef.OFP_NO_BUFFER,
                               out_port = ofdef.OFPP_ANY,
                               out_group = ofdef.OFPG_ANY,
                               priority = ofdef.OFP_DEFAULT_PRIORITY,
                               match = ofdef.ofp_match_oxm(
                                            oxm_fields = [
                                                  ofdef.create_oxm(ofdef.NXM_NX_REG7_W, 0x4000 if islocal else 0, 0x4000),
                                                  match_network(nid),
                                                  ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE, ofdef.ETHERTYPE_ARP),
                                                  ofdef.create_oxm(ofdef.OXM_OF_ARP_TPA, ofdef.ip4_addr(ip)),
                                                  ofdef.create_oxm(ofdef.OXM_OF_ARP_OP, ofdef.ARPOP_REQUEST)]
                                                + ([ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00')]
                                                  if broadcast else [])
                                        ),
                               instructions = create_instructions([ofdef.nx_action_reg_move(
                                                                n_bits = 32,
                                                                src = ofdef.OXM_OF_ARP_SPA,
                                                                dst = ofdef.OXM_OF_ARP_TPA),
                                                           ofdef.nx_action_reg_move(
                                                                n_bits = 48,
                                                                src = ofdef.OXM_OF_ARP_SHA,
                                                                dst = ofdef.OXM_OF_ARP_THA),
                                                           ofdef.nx_action_reg_move(
                                                                n_bits = 48,
                                                                src = ofdef.OXM_OF_ETH_SRC,
                                                                dst = ofdef.OXM_OF_ETH_DST),
                                                           ofdef.ofp_action_set_field(
                                                                  field = ofdef.create_oxm(
                                                                        ofdef.OXM_OF_ARP_SPA,
                                                                        ofdef.ip4_addr(ip)
                                                                    )),
                                                           ofdef.ofp_action_set_field(
                                                                  field = ofdef.create_oxm(
                                                                        ofdef.OXM_OF_ARP_SHA,
                                                                        ofdef.mac_addr(mac),
                                                                    )),
                                                           ofdef.ofp_action_set_field(
                                                                  field = ofdef.create_oxm(
                                                                        ofdef.OXM_OF_ETH_SRC,
                                                                        ofdef.mac_addr(mac),
                                                                    )),
                                                            ofdef.ofp_action_set_field(
                                                                  field = ofdef.create_oxm(
                                                                        ofdef.OXM_OF_ARP_OP,
                                                                        ofdef.ARPOP_REPLY
                                                                    )
                                                                )
                                                           ])
                               )
            def _create_flow2(ip, mac, nid, pid, islocal, broadcast,ifpass):
                if ifpass:
                    ins = [ofdef.ofp_instruction_goto_table(table_id=arp_next)]
                else:
                    ins = [ofdef.ofp_instruction_actions(type=ofdef.OFPIT_CLEAR_ACTIONS)]
                return ofdef.ofp_flow_mod(
                               table_id = arp,
                               cookie = 0x5 | (0x2 if islocal else 0),
                               cookie_mask = 0xffffffffffffffff,
                               command = ofdef.OFPFC_ADD,
                               buffer_id = ofdef.OFP_NO_BUFFER,
                               out_port = ofdef.OFPP_ANY,
                               out_group = ofdef.OFPG_ANY,
                               priority = ofdef.OFP_DEFAULT_PRIORITY + 10,
                               match = ofdef.ofp_match_oxm(
                                            oxm_fields = [
                                                  ofdef.create_oxm(ofdef.OXM_OF_IN_PORT, pid),
                                                  match_network(nid),
                                                  ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE, ofdef.ETHERTYPE_ARP),
                                                  ofdef.create_oxm(ofdef.OXM_OF_ARP_TPA, ofdef.ip4_addr(ip)),
                                                  ofdef.create_oxm(ofdef.OXM_OF_ARP_OP, ofdef.ARPOP_REQUEST)]
                                                + ([ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W, b'\x01\x00\x00\x00\x00\x00', b'\x01\x00\x00\x00\x00\x00')]
                                                  if broadcast else [])
                                        ),
                               instructions=ins
                               )
            for (network, ip, port), value in currentfreearps.items():
                if (network, ip, port) not in lastfreearps or lastfreearps[(network, ip, port)] != value \
                        or lastlogportinfo.get(port) != currentlogportinfo.get(port) \
                        or lastlognetinfo.get(network) != currentlognetinfo.get(network):
                    if port in currentlogportinfo and network in currentlognetinfo:
                        pid, _ = currentlogportinfo[port]
                        nid, _ = currentlognetinfo[network]
                        mac, broadcast, ifpass = value
                        cmds.append(_create_flow2(ip, mac, nid, pid, True, broadcast, ifpass))
            # phynetdict = {}
            # for n in current_arps:
            #     phynet = n.physicalnetwork
            #     phynetdict.setdefault(phynet, []).append(n)
            #===================================================================
            # for p in currentphyportinfo:
            #     if p not in lastphyportinfo or lastphyportinfo[p] != currentphyportinfo[p]:
            #         pid, phynet = currentphyportinfo[p]
            #         if phynet in phynetdict:
            #             for n in phynetdict[phynet]:
            #                 if n in current_arps:
            #                     nid, _ = currentlognetinfo[n]
            #                     for ip, mac, islocal, port, broadcast in current_arps[n]:
            #                         if not islocal and port != p:
            #                             cmds.append(_create_flow(ip, mac, nid, islocal, broadcast))
            #===================================================================
            for n, arps in current_arps.items():
                if n in currentlognetinfo:
                    nid, _ = currentlognetinfo[n]
                    if n not in last_arps or n not in lastlognetinfo or lastlognetinfo[n] != currentlognetinfo[n]:
                        send_arps = arps
                    else:
                        send_arps = {k:v for k,v in arps.items()
                                     if k not in last_arps[n] or last_arps[n] != v}
                    for (ip, islocal), (mac, broadcast) in send_arps.items():
                        cmds.append(_create_flow(ip, mac, nid, islocal, broadcast))
            await self.execute_commands(connection, cmds)
        except Exception:
            self._logger.warning("Unexpected exception in ARPUpdater. Will ignore and continue.", exc_info = True)

                            
        
@defaultconfig
@depend(ofpportmanager.OpenflowPortManager, objectdb.ObjectDB)
class ARPResponder(FlowBase):
    "Send ARP respond"
    _tablerequest = (("arp", ('l2input',), ''),
                     ("l3input", ('arp',), ''),
                     ("l2output", ("l3input",), ''),
                     ("egress", ("l2output",), ''))
    # Prepush ARP entries with Flows, so the switch directly responds an ARP request
    _default_prepush = True
    # When using prepush=True, only responds a broadcast ARP request; let unicast ARP request
    # reach the other side
    _default_broadcastonly = True
    # Drop ARP requests with unknown IP addresses. The ports will not be able to access external IPs,
    # so use with caution.
    _default_disableothers = False
    _default_enable_freearp = True
    _default_enable_freearp_networktypes = ("vlan", "native")

    def __init__(self, server):
        FlowBase.__init__(self, server)
        self.apiroutine = RoutineContainer(self.scheduler)
        self.apiroutine.main = self._main
        self.routines.append(self.apiroutine)
        self._flowupdaters = {}
        self._extra_arps = {}
        self.createAPI(api(self.createproxyarp),
                       api(self.removeproxyarp))

    def createproxyarp(self, connection, arpentries):
        '''
        Create ARP respond flow for specified ARP entries, each is a tuple
        ``(ip_address, mac_address, logical_network_id, local)``. When local is True,
        only respond to ARP request from logical port; when local is False,
        only respond to ARP request from physical port; respond to both else. 
        '''
        arp_list = self._extra_arps.setdefault(connection, [])
        arp_list.extend(arpentries)
        if connection in self._flowupdaters:
            self._flowupdaters[connection].updateobjects([ReferenceObject(LogicalNetwork.default_key(nid)) for _, _, nid, _ in arpentries])

    def removeproxyarp(self, connection, arpentries):
        '''
        Remove specified ARP entries.
        '''
        arp_list = self._extra_arps[connection]
        for entry in arpentries:
            try:
                arp_list.remove(entry)
            except KeyError:
                pass
        if connection in self._flowupdaters:
            self._flowupdaters[connection].updateobjects([ReferenceObject(LogicalNetwork.default_key(nid)) for _, _, nid, _ in arpentries])

    async def _main(self):
        flow_init = FlowInitialize.createMatcher(_ismatch = lambda x: self.vhostbind is None or x.vhost in self.vhostbind)
        conn_down = OpenflowConnectionStateEvent.createMatcher(state = OpenflowConnectionStateEvent.CONNECTION_DOWN,
                                                               _ismatch = lambda x: self.vhostbind is None or x.createby.vhost in self.vhostbind)
        while True:
            ev, m = await M_(flow_init, conn_down)
            if m is flow_init:
                c = ev.connection
                self.apiroutine.subroutine(self._init_conn(c))
            else:
                c = ev.connection
                self.apiroutine.subroutine(self._remove_conn(c))

    async def _init_conn(self, conn):
        # Default
        if conn in self._flowupdaters:
            arpupdater = self._flowupdaters.pop(conn)
            arpupdater.close()
        arpupdater = ARPUpdater(conn, self)
        #flowupdater = VXLANFlowUpdater(conn, self)
        self._flowupdaters[conn] = arpupdater
        arpupdater.start()
        if self.disableothers:
            ofdef = conn.openflowdef
            arptable = self._gettableindex('arp', conn.protocol.vhost)
            await conn.protocol.batch((ofdef.ofp_flow_mod(table_id = arptable,
                                                             command = ofdef.OFPFC_ADD,
                                                             priority = 1,
                                                             buffer_id = ofdef.OFP_NO_BUFFER,
                                                             out_port = ofdef.OFPP_ANY,
                                                             out_group = ofdef.OFPG_ANY,
                                                             match = ofdef.ofp_match_oxm(
                                                                    oxm_fields = [ofdef.create_oxm(ofdef.OXM_OF_ETH_TYPE,
                                                                                                   ofdef.ETHERTYPE_ARP),
                                                                                  ofdef.create_oxm(ofdef.OXM_OF_ARP_OP,
                                                                                                   ofdef.ARPOP_REQUEST)
                                                                                  ] +
                                                                                  ([ofdef.create_oxm(ofdef.OXM_OF_ETH_DST_W,
                                                                                                    b'\x01\x00\x00\x00\x00\x00',
                                                                                                    b'\x01\x00\x00\x00\x00\x00')
                                                                                   ]
                                                                                  if self.broadcastonly else []),
                                                                ),
                                                             instructions = [ofdef.ofp_instruction_actions(type = ofdef.OFPIT_CLEAR_ACTIONS)]
                                                             ),
                                          ), conn, self.apiroutine)

    async def _remove_conn(self, conn):
        # Do not need to modify flows
        if conn in self._flowupdaters:
            self._flowupdaters.pop(conn).close()



'''
Created on 2016/9/13

:author: think
'''

from vlcp.protocol.protocol import Protocol
from vlcp.config import defaultconfig
from vlcp.event import Event, withIndices, ConnectionWriteEvent
from collections import deque
from vlcp.utils.zookeeper import ZooKeeperRequest, ZooKeeperReply, CONNECT_PACKET, HEADER_PACKET,\
        WATCHER_EVENT_XID, ZOO_SYNC_CONNECTED_STATE
import vlcp.utils.zookeeper as zk
import logging
import os
from vlcp.event.connection import ConnectionResetException
from contextlib import closing
from namedstruct import dump
from json import dumps
from vlcp.event.ratelimiter import RateLimiter
from vlcp.event.future import RoutineFuture
from vlcp.event.runnable import RoutineException
from vlcp.event.runnable import _close_generator
from vlcp.event.lock import Lock

@withIndices('state', 'connection', 'connmark', 'createby')
class ZooKeeperConnectionStateEvent(Event):
    UP = 'up'
    DOWN = 'down'
    NOTCONNECTED = 'notconnected'

@withIndices('connection', 'connmark', 'createby')
class ZooKeeperMessageEvent(Event):
    pass

@withIndices('xid')
class ZooKeeperResponseEvent(ZooKeeperMessageEvent):
    pass

@withIndices('type', 'state', 'path')
class ZooKeeperWatcherEvent(ZooKeeperMessageEvent):
    pass

@withIndices()
class ZooKeeperHandshakeEvent(ZooKeeperMessageEvent):
    pass

class ZooKeeperProtocolException(Exception):
    pass

class ZooKeeperException(Exception):
    pass

class ZooKeeperRetryException(ZooKeeperException):
    '''
    Connection lost or not connected on handshake
    '''
    pass

class ZooKeeperSessionExpiredException(ZooKeeperException):
    '''
    Handshake reports the session is expired
    '''
    pass

class ZooKeeperRequestTooLargeException(ZooKeeperException):
    '''
    Request is too large, which may break every thing, so we reject it
    '''
    pass

@withIndices('priority')
class ZooKeeperWriteEvent(ConnectionWriteEvent):
    LOW = 0
    MIDDLE = 1
    HIGH = 2

@defaultconfig
class ZooKeeper(Protocol):
    '''
    ZooKeeper protocol
    '''
    # Usually we should connect to another server, this is done by ZooKeeperClient
    _default_persist = False
    # default ZooKeeper port
    _default_defaultport = 2181
    _default_createqueue = True
    _default_buffersize = 524288
    # Limit the response and watcher queue size
    _default_messagequeuesize = 1024
    # Send ping command when the connection is idle
    _default_keepalivetime = None
    _default_writekeepalivetime = 5
    # Disconnect when the ping command does not receive response
    _default_keepalivetimeout = 10
    _default_connect_timeout = 5
    _default_tcp_nodelay = True
    # Limit the data write queue size
    _default_writequeuesize = 1024
    _logger = logging.getLogger(__name__ + '.ZooKeeper')
    def __init__(self):
        Protocol.__init__(self)

    async def init(self, connection):
        await Protocol.init(self, connection)
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.messagepriority, ZooKeeperConnectionStateEvent.createMatcher(connection = connection), ('connstate', connection)))
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.messagepriority + 1, ZooKeeperMessageEvent.createMatcher(connection = connection), ('message', connection), self.messagequeuesize))
        connection.createdqueues.append(connection.scheduler.queue.addSubQueue(\
                self.messagepriority + 2, ZooKeeperResponseEvent.createMatcher(connection = connection, xid = zk.PING_XID)))
        if hasattr(connection, 'queue'):
            connection.queue.addSubQueue(1, ZooKeeperWriteEvent.createMatcher(connection = connection,
                                                                                        priority = ZooKeeperWriteEvent.MIDDLE),
                                         None, self.writequeuesize)
            connection.queue.addSubQueue(2, ZooKeeperWriteEvent.createMatcher(connection = connection,
                                                                                        priority = ZooKeeperWriteEvent.HIGH),
                                         None, self.writequeuesize)
        # Use limiter to limit the request serialization in one iteration
        connection._rate_limiter = (RateLimiter(0xfffff, connection), RateLimiter(0xfffff, connection))
        await self.reconnect_init(connection)

    async def reconnect_init(self, connection):
        await Protocol.reconnect_init(self, connection)
        connection.xid = ord(os.urandom(1)) + 1
        connection.zookeeper_requests = {}
        connection.zookeeper_handshake = False
        connection.zookeeper_lastzxid = 0
        connection.zookeeper_last_watch_zxid = 0
        await connection.wait_for_send(ZooKeeperConnectionStateEvent(ZooKeeperConnectionStateEvent.UP,
                                                                      connection,
                                                                      connection.connmark,
                                                                      self))

    def serialize(self, connection, event):
        event._zookeeper_sent = True
        return Protocol.serialize(self, connection, event)

    def parse(self, connection, data, laststart):
        events = []
        start = 0
        while True:
            result = ZooKeeperReply.parse(data[start:])
            if result is None:
                break
            reply, size = result
            start += size
            if not connection.zookeeper_handshake:
                reply.zookeeper_type = CONNECT_PACKET
                reply._autosubclass()
                connection.zookeeper_handshake = True
                events.append(ZooKeeperHandshakeEvent(connection, connection.connmark, self, message = reply))
            else:
                reply.zookeeper_type = HEADER_PACKET
                reply._autosubclass()
                if reply.zxid > 0:
                    connection.zookeeper_lastzxid = reply.zxid
                if reply.xid >= 0:
                    xid = reply.xid
                    if xid not in connection.zookeeper_requests:
                        raise ZooKeeperProtocolException('xid does not match: receive %r' % (reply.xid,))
                    request_type = connection.zookeeper_requests.pop(xid)
                    reply.zookeeper_request_type = request_type
                    reply._autosubclass()
                if reply.xid == WATCHER_EVENT_XID:
                    connection.zookeeper_last_watch_zxid = connection.zookeeper_lastzxid
                    reply.last_zxid = connection.zookeeper_lastzxid
                    if reply.state != ZOO_SYNC_CONNECTED_STATE:
                        self._logger.warning("Receive abnormal watch event: %s", dumps(dump(reply, tostr=True)))
                    events.append(ZooKeeperWatcherEvent(connection, connection.connmark,
                                                        self, reply.type, reply.state, b'' if reply.path is None else reply.path,
                                                        message = reply))
                else:
                    events.append(ZooKeeperResponseEvent(connection, connection.connmark,
                                                         self, reply.xid, message = reply))
        if laststart == len(data):
            # Remote write close
            events.append(ConnectionWriteEvent(connection, connection.connmark, data = b'', EOF = True))
        return (events, len(data) - start)

    def _send(self, connection, request, container, priority = 0):
        return self._senddata(connection, request._tobytes(), container, priority)

    async def _senddata(self, connection, data, container = None, priority = 0):
        connwrite = ZooKeeperWriteEvent(connection, connection.connmark, priority, data = data)
        connwrite._zookeeper_sent = False
        try:
            await connection.write(connwrite, False)
        except ConnectionResetException:
            raise ZooKeeperRetryException
        return connwrite

    def _pre_assign_xid(self, connection, request):
        if hasattr(request, 'xid') and request.xid >= 0:
            connection.xid += 1
            if connection.xid > 0x7fffffff:
                connection.xid = 1
            request.xid = connection.xid
        return request.xid

    def _register_xid(self, connection, request):
        if hasattr(request, 'xid') and request.xid >= 0:
            connection.zookeeper_requests[request.xid] = request.type

    async def handshake(self, connection, connrequest, container, extrapackets = []):
        connmark = connection.connmark
        handshake_matcher = ZooKeeperHandshakeEvent.createMatcher(connection, connection.connmark)
        await self._send(connection, connrequest, container, ZooKeeperWriteEvent.HIGH)
        handshake_received = [None]
        if extrapackets:
            def callback(event, matcher):
                handshake_received[0] = event.message
            timeout, r = await container.execute_with_timeout(
                            10,
                            container.with_callback(
                                self.requests(connection, extrapackets, container, priority=ZooKeeperWriteEvent.MIDDLE),
                                callback,
                                handshake_matcher
                            )
                        )
            if timeout:
                raise ZooKeeperRetryException
            receive, lost, retry = r
            if lost or retry:
                if handshake_received[0] is not None:
                    # We have received the handshake packet, but not the extra responses
                    # Maybe the session is expired
                    if handshake_received[0].timeOut <= 0:
                        raise ZooKeeperSessionExpiredException
                raise ZooKeeperRetryException
        else:
            receive = []
        if handshake_received[0] is None:
            if not connection.connected or connection.connmark != connmark:
                raise ZooKeeperRetryException
            else:
                conn_matcher = ZooKeeperConnectionStateEvent.createMatcher(ZooKeeperConnectionStateEvent.DOWN,
                                                                           connection,
                                                                           connmark)
                timeout, ev, m = await container.wait_with_timeout(10, handshake_matcher, conn_matcher)
                if timeout:
                    self._logger.warning('Handshake timeout, connection = %r', connection)
                    raise ZooKeeperRetryException
                elif m is conn_matcher:
                    raise ZooKeeperRetryException
                else:
                    handshake_received[0] = ev.message
                    if handshake_received[0].timeOut <= 0:
                        raise ZooKeeperSessionExpiredException
        return (handshake_received[0], receive)

    async def async_requests(self, connection, requests, container = None, priority = 0):
        '''
        :return: (matchers, sendall), where matchers are event matchers for the requests; sendall
                 is an async function to send to requests. Use `await sendall()` to send the requests.
        '''
        async with Lock((connection, 'async_requests', priority), connection.scheduler):
            matchers = []
            for r in requests:
                xid = self._pre_assign_xid(connection, r)
                resp_matcher = ZooKeeperResponseEvent.createMatcher(connection, connection.connmark, None, xid)
                matchers.append(resp_matcher)
            alldata = []
            for r in requests:
                if priority < ZooKeeperWriteEvent.HIGH:
                    # Test if already limited by consuming 1 byte
                    await connection._rate_limiter[priority].limit(1)
                data = r._tobytes()
                if len(data) >= 0xfffff:
                    # This is the default limit of ZooKeeper, reject this request
                    raise ZooKeeperRequestTooLargeException('The request is %d bytes which is too large for ZooKeeper' % len(data))
                if priority < ZooKeeperWriteEvent.HIGH:
                    await connection._rate_limiter[priority].limit(len(data) - 1)
                alldata.append(data)
            for r in requests:
                self._register_xid(connection, r)
            async def _sendall():
                sent_requests = []
                for data in alldata:
                    try:
                        sent_requests.append(await self._senddata(connection, data, container, priority))
                    except ZooKeeperRetryException:
                        raise ZooKeeperRetryException(sent_requests)
                return sent_requests
            return (matchers, _sendall)

    async def requests(self, connection, requests, container = None, callback = None, priority = 0):
        '''
        Send requests by sequence, return all the results (including the lost ones)
        
        :params connection: ZooKeeper connection
        
        :params requests: a sequence of ZooKeeper requests
        
        :params container: routine container of current routine
        
        :params callback: if not None, `callback(request, response)` is called immediately after
                          each response received
        
        :return: `(responses, lost_responses, retry_requests)`, where responses is a list of responses corresponded
                    to the requests (None if response is not received); lost_responses is a list of requests that are sent
                    but the responses are lost due to connection lost, it is the caller's responsibility to determine whether
                    the call is succeeded or failed; retry_requests are the requests which are not sent and are safe to retry. 
        '''
        matchers, _sendall = await self.async_requests(connection, requests, container, priority)
        requests_dict = dict((m,r) for m,r in zip(matchers, requests))
        connmark = connection.connmark
        if not connection.connected:
            
            return (), (), requests
        conn_matcher = ZooKeeperConnectionStateEvent.createMatcher(ZooKeeperConnectionStateEvent.DOWN,
                                                                   connection,
                                                                   connmark)
        replydict = {}
        if callback:
            def matcher_callback(event, matcher):
                callback(requests_dict[matcher], event.message)
        else:
            matcher_callback = None
        wait_result = RoutineFuture(
                            connection.with_exception(
                                    connection.wait_for_all(
                                        *matchers,
                                        eventdict=replydict,
                                        callback=matcher_callback
                                    ),
                                    conn_matcher
                                ),
                            connection
                        )
        try:
            try:
                sent_events = await _sendall()
            except ZooKeeperRetryException as exc:
                sent_events = exc.args[0]
            try:
                await wait_result
            except RoutineException:
                receive_all = False
            else:
                receive_all = True
            responses = [replydict[m].message if m in replydict else None for m in matchers]
            received_responses = dict((k,v) for k,v in zip(requests, responses) if v is not None)
            if receive_all:
                return (responses, [], [])
            else:
                # Some results are missing
                lost_responses = [r for r,c in zip(requests, sent_events) if c._zookeeper_sent and r not in received_responses]
                retry_requests = [r for r,c in zip(requests, sent_events) if not c._zookeeper_sent] + requests[len(sent_events):]
                return (responses, lost_responses, retry_requests)
        finally:
            wait_result.close()

    async def error(self, connection):
        await connection.wait_for_send(ZooKeeperConnectionStateEvent(ZooKeeperConnectionStateEvent.DOWN,
                                                                      connection,
                                                                      connection.connmark,
                                                                      self))
        await Protocol.error(self, connection)

    async def closed(self, connection):
        await connection.wait_for_send(ZooKeeperConnectionStateEvent(ZooKeeperConnectionStateEvent.DOWN,
                                                                      connection,
                                                                      connection.connmark,
                                                                      self))
        await Protocol.closed(self, connection)

    async def notconnected(self, connection):
        await connection.wait_for_send(ZooKeeperConnectionStateEvent(ZooKeeperConnectionStateEvent.NOTCONNECTED,
                                                                      connection,
                                                                      connection.connmark,
                                                                      self))
        await Protocol.notconnected(self, connection)

    async def keepalive(self, connection):
        try:
            timeout, result = await connection.execute_with_timeout(
                                    self.keepalivetimeout,
                                    self.requests(connection,
                                        [zk.RequestHeader(xid = zk.PING_XID, type = zk.ZOO_PING_OP)],
                                        connection,
                                        priority=ZooKeeperWriteEvent.HIGH)
                                )
            if timeout:
                await connection.reset(True)
            else:
                _, lost, retry = result
                if lost or retry:
                    await connection.reset(True)
        except Exception as exc:
            await connection.reset(True)
            raise exc
